import argparse
import dataclasses
import qrcode
import qrcode.image.svg
import qwif.contents as codes

from pkg_resources import get_distribution
from sys import stderr, exit


def populate_dataclass_parser(
    parser: argparse.ArgumentParser,
    group_name: str = "fields"
):
    """ Populate an argparse.ArgumentParser from a dataclass.

    This expects an argparse.ArgumentParser (typically created with
    `add_parser`), with a default set for 'klass' that is a dataclass.

    Fields from this dataclass are then added as arguments to the parser.
    """

    group = parser.add_argument_group(group_name)

    for field in dataclasses.fields(parser.get_default('klass')):
        default = field.default
        required = False
        action = "store"
        help = f"Set {field.name.replace('_', ' ')} (default: %(default)s)"

        if isinstance(field.default, dataclasses._MISSING_TYPE):
            default = ""
            required = True
            help = f"Set {field.name.replace('_', ' ')} (required)"

        if field.default is True:
            action = "store_false"

        if field.default is False:
            action = "store_true"

        group.add_argument(
            f"--{field.name.replace('_', '-')}",
            help=help,
            default=default,
            required=required,
            action=action
        )

    return parser


def cli():
    """ Get an argparse.ArgumentParser implementing a CLI for qwif.
    """
    parser = argparse.ArgumentParser(
        description="Generate QR code for WIFI and more.",
    )

    parser.add_argument(
        "--version", action='version', version=get_distribution('qwif').version
    )

    # Common is a parent parser for all subparsers that provides common CLI
    # switches/options.

    common = argparse.ArgumentParser(add_help=False)
    common.add_argument(
        "-o", "--output",
        type=argparse.FileType('wb'),
        default="-",
        help="Output file (default: stdout)"
    )

    sub = parser.add_subparsers()

    contact = sub.add_parser(
        'contact',
        help="Encode contact details",
        parents=[common],
        aliases=["c"],
        allow_abbrev=True
    )
    contact.set_defaults(klass=codes.Mecard)
    populate_dataclass_parser(contact)

    event = sub.add_parser(
        'event',
        help="Encode event",
        parents=[common],
        aliases=["e"],
    )

    event.set_defaults(klass=codes.Event)
    populate_dataclass_parser(event)

    geo = sub.add_parser(
        'geo',
        help="Encode geographic location",
        parents=[common],
        aliases=["g"],
    )
    geo.set_defaults(klass=codes.Geo)
    populate_dataclass_parser(geo)

    mail = sub.add_parser(
        'mail',
        help="Encode an email address",
        parents=[common],
        aliases=["m", "email"],
    )
    mail.set_defaults(klass=codes.Email)
    populate_dataclass_parser(mail)

    otp = sub.add_parser(
        'otp',
        help="Encode (T|H)OTP keys",
        parents=[common],
        aliases=["o"],
    )
    otp.set_defaults(klass=codes.OTP)
    populate_dataclass_parser(otp)

    text = sub.add_parser(
        'text',
        help="Encode arbitrary text",
        parents=[common],
        aliases=["t"],
    )
    text.set_defaults(klass=codes.Text)
    populate_dataclass_parser(text)

    url = sub.add_parser(
        'url',
        help="Encode a URL",
        parents=[common],
        aliases=["u"],
    )
    url.set_defaults(klass=codes.Url)
    populate_dataclass_parser(url)

    wifi = sub.add_parser(
        'wifi',
        help="Encode WIFI configuration",
        parents=[common],
        aliases=["w"],
    )
    wifi.set_defaults(klass=codes.Wifi)
    populate_dataclass_parser(wifi)

    return parser


def main(argv: tuple = None):
    parser = cli()
    args = vars(parser.parse_args(argv))

    if not args:
        parser.print_help()
        exit(1)

    output = args.pop('output')

    try:
        code = args.pop('klass')(**args)
        code.validate()
    except ValueError as e:
        print(f"ERROR: {e}", file=stderr)
        exit(1)

    qr = qrcode.QRCode(
        version=1,
        error_correction=qrcode.constants.ERROR_CORRECT_H,
    )

    qr.add_data(code)

    if output.isatty():
        qr.print_tty(out=output)
    else:
        qr.image_factory = qrcode.image.svg.SvgPathImage
        qr.make_image().save(
            output.buffer if hasattr(output, 'buffer') else output
        )



""" A collection of formatters for common QR code types.

This module aims to implement helpers for QR types listed at
https://github.com/zxing/zxing/wiki/Barcode-Contents.
"""

import dataclasses

from dataclasses import dataclass
from datetime import datetime
from os import getpid
from socket import getfqdn
from typing import ClassVar, Dict, List
from urllib.parse import quote


@dataclass
class ContentBase(object):

    special_characters: ClassVar[List[str]] = []
    escape_character: ClassVar[str] = "\\"
    field_names: ClassVar[Dict[str, str]] = dict()
    field_order: ClassVar[List[str]] = []
    prefix: ClassVar[str] = ""

    def validate(self):
        """ Validate field values and contents.
        """
        pass

    def encode(self):
        """ Encode instance data as a string.
        """

        self.validate()

        format_fields = self.field_order

        if not format_fields:
            _fields = self.field_names.keys() or \
                (field.name for field in dataclasses.fields(self))

            format_fields = sorted(_fields)

        fragments = [
            "".join([
                f"{self.field_names.get(field, field)}",
                ":",
                f"{self.escape_string(self.__dict__[field])}"
            ]) for field in filter(
                lambda k: self.__dict__.get(k) not in ["", None],
                format_fields
            )
        ]

        if len(self.prefix):
            return f"{self.prefix}:{';'.join(fragments)};;"

        return f"{';'.join(fragments)};;"

    def __str__(self):
        return self.encode()

    def escape_string(self, data: str):
        """ Escape special characters in a string.
        """

        if not len(self.special_characters):
            return data

        special_charset = set(self.special_characters)

        return data.translate(
            str.maketrans(
                dict(
                    zip(
                        special_charset,
                        [
                            f"{self.escape_character}{char}"
                            for char in special_charset
                        ]
                    )
                )
            )
        )


@dataclass
class Text(ContentBase):
    """ Handle raw text content, with optional escaping.
    """

    text: str

    def encode(self):
        return self.escape_string(self.text or "")


@dataclass
class Email(ContentBase):
    """ Handle mailto strings.
    """

    mailto: str
    subject: str = ""
    cc: str = ""
    bcc: str = ""

    def validate(self):
        if not len(self.mailto):
            raise ValueError("mailto must be a non-empty string")

    def encode(self):

        self.validate()

        params = [
            "=".join(param, quote(self.__dict__[param]))
            for param in filter(
                lambda param: len(self.__dict__[param]),
                ["subject", "cc", "bcc"]
            )
        ]

        if params:
            return f"mailto:{self.mailto}?{'&'.join(params)}"

        return f"mailto:{self.mailto}"


@dataclass
class Telephone(ContentBase):
    number: str

    def validate(self):
        if not len(self.number):
            raise ValueError("number must be a non-empty string")

        if self.number.startswith("00"):
            self.number = "+" + self.number[2:]

        if not self.number[0] == "+":
            raise ValueError(
                "number must be a +-prefixed international dial code"
            )

    def encode(self):
        return f"tel:{self.number}"


@dataclass
class Sms(ContentBase):

    number: str = ""
    message: str = ""

    def encode(self):
        if len(self.message) == 0:
            return f"sms:{self.number}"

        return f"sms:{self.number}:{self.message}"


class Mms(Sms):

    def encode(self):
        if len(self.message) == 0:
            return f"mms:{self.number}"

        return f"mms:{self.number}:{self.message}"


@dataclass
class Url(ContentBase):

    field_names: ClassVar[Dict[str, str]] = dict(
        title="TITLE",
        url="URL"
    )
    url: str
    title: str = ""
    bookmark: bool = False

    def validate(self):
        if not len(self.url):
            raise ValueError("url must be a non-empty string.")

    def encode(self):
        if not self.bookmark:
            return self.url

        return super().encode()


@dataclass
class Wifi(ContentBase):

    # instance vars
    ssid: str
    password: str = ""
    authentication_type: str = "WPA"
    hidden: bool = False
    eap_method: str = "TTLS"
    eap_anonymous_identity: str = ""
    eap_identity: str = ""
    eap_phase_two_method: str = "MSCHAPV2"

    # class vars
    valid_authentication_types: ClassVar[List[str]] = [
        "WPA2-EAP", "WPA", "WEP", "nopass"
    ]
    special_characters: ClassVar[List[str]] = ["\\", ";", ",", ":"]
    escape_character: ClassVar[str] = "\\"

    def validate(self):

        if len(self.ssid) < 1:
            raise ValueError("ssid must be a non-empty string")

        if self.authentication_type not in self.valid_authentication_types:
            raise ValueError(
                f"authentication_type must be one of:"
                f"{','.join(self.valid_authentication_types)}"
            )

        if not self.authentication_type == "nopass":
            if len(self.password) < 1:
                raise ValueError("password must be a non-empty string.")

    def encode(self):
        self.validate()

        fields: Dict[str, str] = dict(
            H=str(self.hidden).lower(),
            S=self.escape_string(self.ssid),
        )

        if self.authentication_type != "nopass":
            fields.update(
                dict(
                    P=self.escape_string(self.password),
                    T=self.authentication_type
                )
            )

        if self.authentication_type == "WPA2-EAP":
            fields.update(
                dict(
                    A=self.escape_string(self.eap_anonymous_identity),
                    E=self.eap_method,
                    I=self.escape_string(self.eap_identity),
                    PH2=self.eap_phase_two_method,
                )
            )

        return "".join(
            [
                "WIFI:",
                ';'.join(
                    [
                        f"{field}:{fields[field]}"
                        for field in sorted(fields.keys())
                    ]
                ),
                ";",
                ";"
            ]
        )


@dataclass
class Mecard(ContentBase):
    name: str
    address: str = ""
    birthday: str = ""
    email: str = ""
    memo: str = ""
    nickname: str = ""
    number: str = ""
    url: str = ""

    #
    prefix: ClassVar[str] = "MECARD"
    field_names: ClassVar[Dict[str, str]] = dict(
        address="ADR",
        birthday="BDAY",
        email="EMAIL",
        memo="NOTE",
        name="N",
        nickname="NICKNAME",
        number="TEL",
        url="URL",
    )

    def validate(self):
        if not len(self.name):
            raise ValueError("name must be a non-empty string.")


@dataclass
class Geo(ContentBase):
    latitude: float
    longitude: float
    elevation: float = 0.0

    prefix: ClassVar[str] = "geo"

    def encode(self):
        return str(
            f"{self.prefix}:"
            f"{self.latitude},{self.longitude},{self.elevation}"
        )


@dataclass
class Event(ContentBase):
    #
    dtformat: ClassVar[str] = "%Y%m%dT%H%M%SZ"

    # Required instance fields
    summary: str
    dtstart: str
    # Optional fields.
    # Not _all_ fields from vevent are supported.
    # * class is not supported
    # * last-mod is not supported
    # * x-prop, iana-prop are not supported.
    categories: str = ""
    comment: str = ""
    contact: str = ""
    created: str = datetime.utcnow().strftime("%Y%m%dT%H%M%SZ")
    description: str = ""
    duration: str = ""
    geo: str = ""
    location: str = ""
    organizer: str = ""
    priority: str = ""
    seq: str = ""
    status: str = ""
    dtstamp: str = datetime.utcnow().strftime("%Y%m%dT%H%M%SZ")
    dtend: str = ""
    uid: str = str(
        f"{datetime.utcnow().strftime('%Y%m%dT%H%M%SZ')}"
        f"-{getpid()}@{getfqdn()}"
    )
    url: str = ""

    def validate(self):

        if self.dtstart == "":
            raise ValueError("dstart is required")

        if "" in [self.dtstart, self.uid, self.dtstamp]:
            raise ValueError("dstart, dstamp and uid are required")

        if self.dtend == "" and self.duration == "":
            raise ValueError("dtend or duration must be provided")

    def encode(self):

        self.validate()

        format_fields = self.field_order

        if not format_fields:
            format_fields = sorted(
                self.field_names.keys() or
                (field.name for field in dataclasses.fields(self))
            )

        fragments = [
            "".join([
                f"{self.field_names.get(field, field).upper()}",
                ":",
                f"{self.escape_string(self.__dict__[field])}"
            ]) for field in filter(
                lambda k: self.__dict__.get(k) not in ["", None],
                format_fields
            )
        ]

        return "\r\n".join(
            ["BEGIN:VEVENT", "\r\n".join(fragments), "END:VEVENT"]
        )


@dataclass
class OTP(ContentBase):
    account_name: str
    secret: str
    issuer: str
    algorithm: str = "sha256"
    type: str = "totp"
    counter: int = 0
    period: int = 30
    digits: int = 6

    def validate(self):

        if self.type not in ["totp", "hotp"]:
            raise ValueError("type must be one of totp, hotp")

        if self.algorithm.upper() not in ["SHA1", "SHA256", "SHA512"]:
            raise ValueError("algorithm must be one of SHA1, SHA256, SHA512")

    def encode(self):
        self.validate()

        uri = str(
            f"optauth://{self.type}/"
            f"{quote(self.issuer)}:{self.account_name}/?"
            f"secret={self.secret}&"
            f"issuer={quote(self.issuer)}&"
            f"algorithm={self.algorithm.upper()}&"
            f"digits={self.digits}"
        )

        if self.type == "hotp":
            return str(f"{uri}&counter={self.counter}")

        return str(f"{uri}&period={self.period}")



import dataclasses
import pytest

from qwif import contents


def test_format_base_escape_string():

    base = contents.ContentBase()

    base.escape_character = "\\"

    # The order in which special characters are defined should be irrelevant.
    # It should specifically not result escape characters getting escaped.
    base.special_characters = ["\\", ";"]
    assert base.escape_string("\\;") == "\\\\\\;"
    base.special_characters = [";", "\\"]
    assert base.escape_string("\\;") == "\\\\\\;"
    assert base.escape_string(";;") == "\\;\\;"


def test_format_base_encode():

    # ContentBase should expose any field keys as format fields by default.
    base_cls = dataclasses.make_dataclass(
        'ContentTest',
        fields=[
            ('test', str, dataclasses.field(default="true")),
            ('time', int, dataclasses.field(default=0)),
        ],
        bases=(contents.ContentBase,)
    )
    base = base_cls()
    assert str(base) == "test:true;time:0;;"

    # fields with a value of `None` should not be present in encoded output
    base = base_cls()
    base.__dict__["time"] = None
    assert str(base) == "test:true;;"

    # If field_names are set, only those should be present in encoded output.
    # Additionally, instance fields names should be mapped to the proper target
    # name specified in `field_names`
    base = base_cls()
    base.field_names = dict(time="T")
    assert str(base) == "T:0;;"
    base.field_names["test"] = "t"
    assert str(base) == "t:true;T:0;;"

    # Setting field_order should ensure only specified fields are present in
    # encoded output, and that fields are listed in the stated order.
    base = base_cls()
    base.field_order = ["time", "test"]
    assert str(base) == "time:0;test:true;;"
    base.__dict__["time"] = None
    assert str(base) == "test:true;;"

    # Setting a `prefix` should result in that prefix being prepended to the
    # encoded output
    base = base_cls()
    base.prefix = "RES"

    assert str(base) == "RES:test:true;time:0;;"


def test_url():

    url = contents.Url(url="https://test")
    assert str(url) == "https://test"
    url.bookmark = True
    assert str(url) == "URL:https://test;;"
    url.title = "test"
    assert str(url) == "TITLE:test;URL:https://test;;"


def test_email():

    email = contents.Email("test@test.test")
    assert str(email) == "mailto:test@test.test"

    email = contents.Email("test@test.test?subject=Meeting&cc=info@test.test")
    assert str(email) == str(
        "mailto:test@test.test?subject=Meeting&cc=info@test.test"
    )


def test_text():

    text = contents.Text("test")
    assert str(text) == "test"


def test_sms():

    sms = contents.Sms(number="1111")
    assert str(sms) == "sms:1111"
    sms.message = "hello"
    assert str(sms) == "sms:1111:hello"
    sms.number = ""
    assert str(sms) == "sms::hello"


def test_mms():
    mms = contents.Mms(number="1111")
    assert str(mms) == "mms:1111"
    mms.message = "hello"
    assert str(mms) == "mms:1111:hello"
    mms.number = ""
    assert str(mms) == "mms::hello"


def test_wifi():

    wifi = contents.Wifi(ssid="test", password="test")
    assert str(wifi) == "WIFI:H:false;P:test;S:test;T:WPA;;"
    wifi.authentication_type = "WEP"
    assert str(wifi) == "WIFI:H:false;P:test;S:test;T:WEP;;"
    wifi.authentication_type = "nopass"
    assert str(wifi) == "WIFI:H:false;S:test;;"
    wifi.authentication_type = "WPA2-EAP"
    wifi.eap_anonymous_identity = "anon"
    wifi.eap_identity = "user"
    assert str(wifi) == str(
        "WIFI:A:anon;E:TTLS;H:false;I:user;P:test;PH2:MSCHAPV2;"
        "S:test;T:WPA2-EAP;;"
    )


def test_otp():
    otp = contents.OTP(
        secret="ORSXG5AK", issuer="test issuer", account_name="test"
    )

    assert str(otp) == str(
        "optauth://totp/test%20issuer:test/?secret=ORSXG5AK"
        "&issuer=test%20issuer&algorithm=SHA256&digits=6&period=30"
    )
    otp.type = "hotp"
    assert str(otp) == str(
        "optauth://hotp/test%20issuer:test/?secret=ORSXG5AK"
        "&issuer=test%20issuer&algorithm=SHA256&digits=6&counter=0"
    )

    with pytest.raises(ValueError):
        otp.type = "whatever"
        otp.validate()

    otp.type = "totp"

    with pytest.raises(ValueError):
        otp.algorithm = "whatever"
        otp.validate()


def test_mecard():

    mecard = contents.Mecard(name="test")
    assert str(mecard) == "MECARD:N:test;;"
    mecard.email = "test@test.test"
    assert str(mecard) == "MECARD:EMAIL:test@test.test;N:test;;"

    with pytest.raises(ValueError):
        mecard.name = ""
        mecard.validate()


def test_event():

    event = contents.Event(
        summary="test event",
        dtstart="00",
        dtend="01"
    )

    assert str(event).startswith("BEGIN:VEVENT")
    assert str(event).endswith("END:VEVENT")



import pytest
import re

from qwif import cli, contents
from sys import stdout


def test_wifi_subcommand():

    args = ("wifi", "--ssid", "test", "--password", "test", "--hidden",)
    data = cli.cli().parse_args(args)

    assert data.klass == contents.Wifi
    assert data.output == stdout

    #
    assert data.hidden is True
    assert data.password == "test"
    assert data.ssid == "test"


def test_otp_subcommand():

    args = (
        "otp", "--secret", "test", "--issuer", "test", "--account-name", "test"
    )

    data = cli.cli().parse_args(args)

    assert data.klass == contents.OTP
    assert data.output == stdout

    #
    assert data.secret == "test"
    assert data.issuer == "test"
    assert data.account_name == "test"


def test_main(capsys):

    # with no subcommand, qwif CLI should print help, then exit(1).
    with pytest.raises(SystemExit):
        cli.main()

    output = capsys.readouterr()
    assert cli.cli().format_usage() in output.err

    # all subcommands have at least one required field, and should have their
    # own help output.
    for subcommand in ["contact", "geo", "mail", "otp", "text", "url", "wifi"]:
        with pytest.raises(SystemExit):
            cli.main((subcommand, ))
        assert subcommand in capsys.readouterr().err

    capsys.readouterr()
    # '--version' should be supported
    with pytest.raises(SystemExit):
        cli.main(("--version", ))
    output = capsys.readouterr()
    assert re.match(r"\d+\.\d+\.\d+", output.out) is not None
